# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"gdown_unzip": "06_uwb.ipynb",
         "pose_dis": "06_uwb.ipynb",
         "waypoint": "06_uwb.ipynb",
         "camera_projection": "01_camera_projection.ipynb",
         "get_images_from_video": "02_video2picture.ipynb",
         "robot_model_loader": "03_robot_model.ipynb",
         "get_key": "04_get_ip.ipynb",
         "myip": "04_get_ip.ipynb",
         "whoami": "04_get_ip.ipynb",
         "get_xbee_address": "04_get_ip.ipynb",
         "get_xbee_address_boat": "04_get_ip.ipynb",
         "find_duckiepond_devices_yaml": "04_get_ip.ipynb",
         "dp_load_config": "04_get_ip.ipynb",
         "dp_get_devices": "04_get_ip.ipynb",
         "device_get_sensors": "04_get_ip.ipynb",
         "sensor_get_topic": "04_get_ip.ipynb",
         "ssh_ping_nano": "04_get_ip.ipynb",
         "ssh_ping_rpi": "04_get_ip.ipynb",
         "test_ssh_intranet": "04_get_ip.ipynb",
         "ssh_connection": "04_get_ip.ipynb",
         "test_ssh": "04_get_ip.ipynb",
         "ip_connection": "04_get_ip.ipynb",
         "test_ping": "04_get_ip.ipynb",
         "ssh_rostopic": "04_get_ip.ipynb",
         "test_rostopic": "04_get_ip.ipynb",
         "random_generate": "05_random_map.ipynb",
         "sub_random_generate": "05_random_map.ipynb",
         "draw_line": "05_random_map.ipynb",
         "dist": "07_tsp.ipynb",
         "distanceGenerate": "07_tsp.ipynb",
         "sortWaypoint": "07_tsp.ipynb",
         "solve_tsp_nearest_neighbor": "07_tsp.ipynb",
         "solve_tsp_held_karp": "07_tsp.ipynb",
         "ros_socket": "08_websocket_rosbridge.ipynb"}

modules = ["utils.py",
           "camera_projection.py",
           "video2picture.py",
           "robot_model.py",
           "get_ip.py",
           "random_map.py",
           "uwb.py",
           "tsp.py",
           "websocket_rosbridge.py"]

doc_url = "https://ARG-NCTU.github.io/arg_robotics_tools/"

git_url = "https://github.com/ARG-NCTU/arg_robotics_tools/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 02_video2picture.ipynb (unless otherwise specified).

__all__ = ['get_images_from_video']

# Cell
import gdown
from zipfile import ZipFile
from PIL import Image
import sys
import os
import cv2

# Cell
def get_images_from_video(video_name, time_F):
    '''
    open and read video,then save the images of video depending on the parameter(time_F) you setup.
    '''
    video_images = []
    vc = cv2.VideoCapture(video_name)
    c = 1

    if vc.isOpened():
        rval, video_frame = vc.read()
    else:
        rval = False

    while rval:
        rval, video_frame = vc.read()

        if(c % time_F == 0):
            video_images.append(video_frame)
        c = c + 1
    vc.release()

    return video_images


# AUTOGENERATED! DO NOT EDIT! File to edit: 06_uwb.ipynb (unless otherwise specified).

__all__ = ['gdown_unzip', 'pose_dis', 'waypoint']

# Cell

import os
import sys
import gdown
import copy
import math
from zipfile import ZipFile

def gdown_unzip(id, filename):
    """download a zipfile and unzip it
    """
    dataset_url = 'https://drive.google.com/u/1/uc?id=' + id
    dataset_name = filename

    if not os.path.isdir(dataset_name):
        gdown.download(dataset_url, output = dataset_name + '.zip', quiet=False)
        zip_file = ZipFile( dataset_name + '.zip')
        #zip_file.extractall()
        zip_file.extractall() # depends on how to zip it
        zip_file.close()


def pose_dis(pose_1, pose_2):
    """Compute distance between pose_1 and pose_2
    """
    x = pose_1[0] - pose_2[0]
    y = pose_1[1] - pose_2[1]
    z = pose_1[2] - pose_2[2]

    dis = math.sqrt(x**2+y**2+z**2)

    return dis

def waypoint(current_pose, Target_pose):
    """Generate a list of way points from current pose to target pose

    Input : current pose, target pose : list [x_pos, y_pos, z_pos, x_ori, y_ori, z_ori, w_ori]
    Return : a list of way points

    """
    waypoint_list = []
    factor = 0.5
    sub_pose = copy.deepcopy(current_pose)

    # threshold : distance between sub_pose and target_pose = 0.05 meter
    dis = pose_dis(sub_pose, Target_pose)
    while dis > 0.05:
        sub_pose[0] = (sub_pose[0] + Target_pose[0])*factor
        sub_pose[1] = (sub_pose[1] + Target_pose[1])*factor
        sub_pose[2] = (sub_pose[2] + Target_pose[2])*factor
        sub_pose[3] = Target_pose[3]
        sub_pose[4] = Target_pose[4]
        sub_pose[5] = Target_pose[5]
        sub_pose[6] = Target_pose[6]

        dis = pose_dis(sub_pose, Target_pose)

        waypoint_list.append(copy.deepcopy(sub_pose))

    waypoint_list.append(Target_pose)

    return waypoint_list


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_camera_projection.ipynb (unless otherwise specified).

__all__ = ['camera_projection']

# Cell

import numpy as np
import scipy as sp
import cv2
from cv2 import aruco
import apriltag
import time
import yaml

import pytransform3d.rotations as pr
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import matplotlib as mpl

import os
import sys
import gdown
from zipfile import ZipFile

from scipy.spatial.transform import Rotation as R
from numpy.linalg import inv

# Cell

class camera_projection:
    def __init__(self):
        """init a camera projection object with default arguments
        """
        self.camera_info_path = 'ViperX_apriltags/camera_info.yaml'
        self.img_path = 'ViperX_apriltags/rgb/'
        self.depth_path = 'ViperX_apriltags/depth/'
        self.tag_size = 0.0415
        self.s = 0.5 * self.tag_size

    def read_camera_info(self):
        """read camera info from yaml file, path is given, camera info contains camera matrix and dist coefts
        """
        with open(self.camera_info_path, "r") as stream:
            try:
                camera_data = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
        self.camera_matrix = np.array(camera_data['camera_matrix']['data'])
        self.camera_matrix = self.camera_matrix.reshape(3, 3)
        self.dist_coeffs = np.array(camera_data['distortion_coefficients']['data'])
        self.dist_coeffs = self.dist_coeffs.reshape(1, 5)
        self.cameraParams_Intrinsic = [self.camera_matrix[0,0], self.camera_matrix[1,1],
                                       self.camera_matrix[0,2], self.camera_matrix[1,2]]

    def read_images(self, idx):
        """this function will load an image depend on the id number, often used in a for loop
        """
        self.img_path = self.img_path + str(idx) + '.png'
        self.depth_path = self.depth_path + str(idx) + '.png'
        self.img = cv2.imread(self.img_path)
        self.gray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
        self.depth = cv2.imread(self.depth_path, -cv2.IMREAD_ANYDEPTH)
        self.img_dst = cv2.cvtColor(self.img, cv2.COLOR_BGR2RGB)

    def apriltag_detection(self):
        """detect if there is apriltag or not in self.gray, the self image
        """
        print("[INFO] detecting AprilTags...")
        options = apriltag.DetectorOptions(families="tag36h11")
        detector = apriltag.Detector(options)
        #results = detector.detect(gray)
        self.detection_results, dimg = detector.detect(self.gray, return_image=True)
        print("[INFO] {} total AprilTags detected".format(len(self.detection_results)))

    def solvePnP(self):
        """this function will output the rotation matrix r_vec and translation matrix t_vex, this two matrixs is important for projection
        """
        img_pts = self.detection_results[0].corners.reshape(1,4,2)
        obj_pt1 = [-self.s, -self.s, 0.0]
        obj_pt2 = [ self.s, -self.s, 0.0]
        obj_pt3 = [ self.s,  self.s, 0.0]
        obj_pt4 = [-self.s,  self.s, 0.0]
        obj_pts = obj_pt1 + obj_pt2 + obj_pt3 + obj_pt4
        obj_pts = np.array(obj_pts).reshape(4,3)

        _, self.r_vec, self.t_vec = cv2.solvePnP(obj_pts, img_pts, self.camera_matrix,
                                       self.dist_coeffs, flags=cv2.SOLVEPNP_ITERATIVE)
        R_mat, _ = cv2.Rodrigues(self.r_vec)
        T = np.hstack((R_mat, self.t_vec)).reshape(3,4)
        tag_pose = np.vstack((T, [0,0,0,1])).reshape(4,4)
        dist = np.linalg.norm(self.t_vec)

    def draw_point(self, tag_2_inv, base2joint):
        """for visualization, draw the project points on the image is important
        """
        # --------------- project a point ---------------
        tag2joint = np.matmul(tag_2_inv, base2joint)
        obj_pts = np.array([tag2joint[0,3], tag2joint[1,3], tag2joint[2,3]]).reshape(1,3)
        proj_img_pts, jac = cv2.projectPoints(obj_pts, self.r_vec, self.t_vec,
                                              self.camera_matrix, self.dist_coeffs)
        proj_img_pts = np.array(proj_img_pts).reshape(2,1)
        # --------------- draw a point ---------------
        draw_image = cv2.circle(self.img_dst, (int(proj_img_pts[0]), int(proj_img_pts[1])),
                                radius=5, color=(255, 0, 0), thickness=-1)
        return draw_image


# AUTOGENERATED! DO NOT EDIT! File to edit: 05_random_map.ipynb (unless otherwise specified).

__all__ = ['random_generate', 'sub_random_generate', 'draw_line']

# Cell
from random import uniform
import matplotlib.pyplot as plt
import numpy as np

def random_generate(offset_x, offset_y, point_num, area_length_x, area_length_y):
    """
        generate random points with mutiple parameters.
        Args:
            offset_x : offset for x's range.
            offset_y : offset for y's range.
            point_num : number of points to generate.
            area_length_x : x's range for generate random points.
            area_length_y : y's range for generate random points.
        Returns:
            rand_point_x : list of random points' x coordinate.
            rand_point_y : list of random points' y coordinate.
    """
    rand_point_x = []
    rand_point_y = []
    rand_point_x = [(uniform(-area_length_x/2, area_length_x/2) + offset_x) for _ in range(point_num)]
    rand_point_y = [(uniform(-area_length_y/2, area_length_y/2) + offset_y) for _ in range(point_num)]
    return rand_point_x, rand_point_y

def sub_random_generate(offset_x, offset_y, point_num, area_length_x, area_length_y, sub_area_num, sub_offset_x, sub_offset_y):
    """
        generate random points with mutiple parameters, including sub_area_num and sub_offset.
        Args :
            offset_x : offset for x's range.
            offset_y : offset for y's range.
            point_num : number of points to generate.
            area_length_x : x's range for generate random points.
            area_length_y : y's range for generate random points.
            sub_area_num :
            sub_offset_x :
            sub_offset_y :
        Returns :
            rand_point_x : list of random points' x coordinate.
            rand_point_y : list of random points' y coordinate.
        Outputs :
            output_point.txt : random points' x and y coordinate.
    """
    rand_point_x = []
    rand_point_y = []
    path = "output_point.txt"
    f = open(path, "w")
    for i in range(sub_area_num):
        generated_x = [(uniform(-area_length_x/2 + i*(area_length_x/sub_area_num)+sub_offset_x, -area_length_x/2 + (i+1)*(area_length_x/sub_area_num)-sub_offset_x)
            + offset_x) for _ in range(int(point_num/sub_area_num))]
        generated_y = [(uniform(-area_length_y/2+sub_offset_y, area_length_y/2-sub_offset_y) + offset_y) for _ in range(int(point_num/sub_area_num))]
        rand_point_x.extend(generated_x)
        rand_point_y.extend(generated_y)
        print("Area", i, ":", file=f)
        for j in range(int(point_num/sub_area_num)):
            print(generated_x[j], generated_y[j], file=f)
    f.close()
    return rand_point_x, rand_point_y

def draw_line(sub_area_num, x_min, x_increment, y_min, y_max):
    """
    draw straight line in a plot.
    Args:
        sub_area_num : number of area that separated by lines.
        x_min : where the line starts.
        x_increment : distance of two lines.
        y_min : line's y lower coordinate.
        y_max : line's y upper coordinate.
    """

    for i in range(sub_area_num):
        plt.vlines(x_min, y_min, y_max, color='green')
        x_min += x_increment




# AUTOGENERATED! DO NOT EDIT! File to edit: 07_tsp.ipynb (unless otherwise specified).

__all__ = ['dist', 'distanceGenerate', 'sortWaypoint', 'solve_tsp_nearest_neighbor', 'solve_tsp_held_karp']

# Cell

import sys
import itertools
import random
import time
import matplotlib.pyplot as plt
import numpy as np
import math
from python_tsp.heuristics import solve_tsp_simulated_annealing

def dist(p1, p2):
    """
        calculate the distance between two waypoints.
        Args:
            p1 : point's (x,y) position.
            p2 : point's (x,y) position.
        Returns:
            distance between two points.
    """
    return math.sqrt(((p1-p2)**2).sum())

def distanceGenerate(point_set):
    """
        generate a distance matrix based on the waypoints.
        Args:
            point_set : a set that contains all waypoint.
        Returns:
            a square matrix, which shows the distance between pairs of waypoint.
    """
    return np.asarray([[dist(np.array(p1), np.array(p2)) for p2 in point_set] for p1 in point_set])

def sortWaypoint(permutation, point_set):
    """
        accoriding to permutation calculated by tsp solver, return new set of waypoint which is sorted.
        Args:
            permutation : the order of waypoints calculated by tsp solver.
            point_set : a set that contains all waypoint.
        Returns:
            new set of waypoint which is sorted.
    """
    return [x for _, x in sorted(zip(permutation, point_set))]

def solve_tsp_nearest_neighbor(distance_matrix):
    """
        calculate tsp problem based on nearest neighbor, an algorithm that solves tsp using greedy assumption.
        Args:
            distance_matrix : a square matrix, which shows the distance between pairs of waypoint.
        Returns:
            A tuple, (path, cost)
            cost : optimal cost of tsp
            path : a orderd list of waypoint index based on distance matrix.
    """
    path = [0]
    cost = 0
    N = distance_matrix.shape[0]
    mask = np.ones(N, dtype=bool)
    mask[0] = False

    for i in range(N-1):
        last = path[-1]
        next_ind = np.argmin(distance_matrix[last][mask]) # find minimum of remaining locations
        next_loc = np.arange(N)[mask][next_ind] # convert to original location
        path.append(next_loc)
        mask[next_loc] = False
        cost += distance_matrix[last, next_loc]
        if(i == N-2):
            cost += distance_matrix[next_loc, 0]

    return path, cost

def solve_tsp_held_karp(distance_matrix):
    """
        calculate tsp problem based on Held-Karp, an algorithm that solves tsp using dynamic programming with memoization.
        Args:
            distance_matrix : a square matrix, which shows the distance between pairs of waypoint.
        Returns:
            A tuple, (path, cost)
            cost : optimal cost of tsp
            path : a orderd list of waypoint index based on distance matrix.
    """
    n = len(distance_matrix)
    C = {}

    # Set transition cost from initial state
    for k in range(1, n):
        C[(1 << k, k)] = (distance_matrix[0][k], 0)

    # Iterate subsets of increasing length and store intermediate results
    # in classic dynamic programming manner
    for subset_size in range(2, n):
        for subset in itertools.combinations(range(1, n), subset_size):
            bits = 0
            for bit in subset:
                bits |= 1 << bit
            # Find the lowest cost to get to this subset
            for k in subset:
                prev = bits & ~(1 << k)
                res = []
                for m in subset:
                    if m == 0 or m == k:
                        continue
                    res.append((C[(prev, m)][0] + distance_matrix[m][k], m))
                C[(bits, k)] = min(res)
    bits = (2**n - 1) - 1

    # Calculate optimal cost
    res = []
    for k in range(1, n):
        res.append((C[(bits, k)][0] + distance_matrix[k][0], k))
    opt, parent = min(res)

    # Backtrack to find full path
    path = []
    for i in range(n - 1):
        path.append(parent)
        new_bits = bits & ~(1 << parent)
        _, parent = C[(bits, parent)]
        bits = new_bits

    # Add implicit start state
    path.append(0)

    return list(reversed(path)), opt




# AUTOGENERATED! DO NOT EDIT! File to edit: 04_get_ip.ipynb (unless otherwise specified).

__all__ = ['get_key', 'myip', 'whoami', 'get_xbee_address', 'get_xbee_address_boat', 'find_duckiepond_devices_yaml',
           'dp_load_config', 'dp_get_devices', 'device_get_sensors', 'sensor_get_topic', 'ssh_ping_nano',
           'ssh_ping_rpi', 'test_ssh_intranet', 'ssh_connection', 'test_ssh', 'ip_connection', 'test_ping',
           'ssh_rostopic', 'test_rostopic']

# Cell
import subprocess
import os
import yaml
import re

def get_key(dict,value):

    for k, v in dict.items():
        for v,v1 in v.items():
            for v1,v2 in v1.items():
                if v2 == value:
                    return k,v

def myip():

    ret_byte = subprocess.check_output(['ifconfig'])
    ret_str = ret_byte.decode('utf-8')
    # Cut string from 'equal symbol' to 'degree C symbol', then convert to float
    en = ret_str[ret_str.find('eno1:'): ret_str.find('lo')]
    ip = en[en.find('inet')+5: en.find('netmask')-2]
    return ip

def whoami(data):
    ip = myip(data)
    machine,device = get_key(data,ip)
    return machine,device

def get_xbee_address(dict,value):
    for k, v in dict.items():
        for v,v1 in v.items():
            for v1,v2 in v1.items():
                if v2 == value:
                    address = dict[k]["rpi"]['xbee_rx']
                    return address

def get_xbee_address_boat(dict,value):
    pair_device = dict[value]["xbee"]["xbee_pair"]
    address = dict[pair_device]["rpi_2"]["xbee_rx"]
    return address

def find_duckiepond_devices_yaml(yaml_filename="duckiepond-devices.yaml"):
    dp_yaml_path = ""
    for root, dirs, files in os.walk(os.path.expanduser('~')):
        for name in files:
            if name == yaml_filename:
                dp_yaml_path = os.path.abspath(os.path.join(root, name))
                break
    return dp_yaml_path

def dp_load_config(dp_yaml_path):
    dp_dict = {}
    with open(dp_yaml_path, 'r') as stream:
        try:
            dp_dict = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    return dp_dict

def dp_get_devices(dp_yaml_path, pattern='boat*'):
    dp_dict = dp_load_config(dp_yaml_path)
    devices = []
    for key in dp_dict.keys():
        match = re.match(pattern, key)
        if match:
            devices.append(key)
    return devices

def device_get_sensors(dict, device='sensor1'):
    sensors = []
    for key in dp_dict[device]['topics'].keys():
        sensors.append(key)
    return sensors

def sensor_get_topic(dp_dict, device='sensor1', find='zed'):
    topic = dp_dict[device]['topics'][find]
    return topic

#ssh functions will not give testing example for now on, since hostname will depend on your running machine

def ssh_ping_nano(hostname):
    response = os.system("ssh $USER@" + hostname + " ping -c 1 192.168.0.100")
    return response

def ssh_ping_rpi(hostname):
    response = os.system("ssh $USER@" + hostname + " ping -c 1 192.168.0.101")
    return response

def test_ssh_intranet():
    error = []
    for ip in hostnames:
        num = ssh_ping_rpi(ip)
        if num!=0:
            error.append("ssh ping rpi error " + ip)
        num = ssh_ping_nano(ip)
        if num!=0:
            error.append("ssh ping nano error " + ip)
    assert not error, "errors occured:\n{}".format("\n".join(error))

def ssh_connection(hostname):
    response = os.system("ssh $USER@" + hostname + " date")
    return response

def test_ssh():
    error = []
    for ip in hostnames:
        num = ssh_connection(ip)
        if num != 0:
            error.append("ssh error " + ip)
    assert not error, "errors occured:\n{}".format("\n".join(error))

def ip_connection(hostname):
    response = os.system("ping -c 1 " + hostname)
    return response

def test_ping():
    error = []
    for ip in hostnames:
        num = ip_connection(ip)
        if num != 0:
            error.append("Network Error " + ip)
    assert not error, "errors occured:\n{}".format("\n".join(error))

def ssh_rostopic(hostname, rosversion="melodic"):
    response = os.system('ssh $USER@' + hostname + ' "source /opt/ros/"' + rosversion + '"/setup.bash && rostopic list"')
    return response

def test_rostopic():
    error = []
    for ip in hostnames:
        num = ssh_rostopic(ip)
        if num != 0:
            error.append("ssh rostopic list error " + ip)
    assert not error, "errors occured:\n{}".format("\n".join(error))



# AUTOGENERATED! DO NOT EDIT! File to edit: 08_websocket_rosbridge.ipynb (unless otherwise specified).

__all__ = ['ros_socket']

# Cell
import roslibpy
import time

# Cell
class ros_socket():
    def __init__(self, ip, port=9090):
        '''
        __init__

        Input:
            ip(type: string) ip address you want to connect
            port(type: int) default is 9090
        '''
        self.ip = ip
        self.port = port
        self.topic = []
        self.node = []
        self.topic_name = ''
        self.topic_type = ''
        self.client = roslibpy.Ros(host = ip, port = port)
        self.client.run()

    def get_topic(self):
        self.topic = self.client.get_topics()
        return self.topic

    def get_node(self):
        self.node = self.client.get_nodes()
        return self.node

    def check_connecting(self):
        print('Is ROS connected?',  self.client.is_connected)

    def subscriber(self, topic_name, subscribe_callback, rate_in_ms=1000):
        '''
        subscriber
        subscribe topic with rate (default = 1sec)

        Input:
            topic_name(type:)
            subscribe_callback(message)
            (type: function) *only one argument message-> that will load with data you subscribing
        '''
        self.topic_name = topic_name
        self.topic_type = self.client.get_topic_type(self.topic_name)
        listener = roslibpy.Topic(self.client, self.topic_name, self.topic_type,throttle_rate = rate_in_ms)
        listener.subscribe(subscribe_callback)

    def publisher(self, topic_name, topic_type, message_data):
        '''
        publisher
        publish message_data to topic_name

        Input:
            topic_name(type:string)
            topic_type(type:)
            message_data(type:topic_type)
        '''
        talker = roslibpy.Topic(client, topic_name, message_data)
        talker.publish(roslibpy.Message({'': message_data}))


    def println(self, ros_list):
        if len(ros_list) == 0:
            print('Empty')
        else:
            for i in ros_list:
                print(i)




# AUTOGENERATED! DO NOT EDIT! File to edit: 03_robot_model.ipynb (unless otherwise specified).

__all__ = ['robot_model_loader']

# Cell
#nbdev_comment from __future__ import print_function

import gdown
from zipfile import ZipFile
import xml.etree.cElementTree as ET
from urdfpy import URDF
import os

class robot_model_loader:
    def __init__(self, url, name):
        self.url = url
        self.name = name

    def load(self):
        """
            download a zipfile and unzip it under data directory
        """
        dataset_url = 'https://drive.google.com/u/1/uc?id=' + self.url
        dataset_name = self.name

        gdown.download(dataset_url, output=dataset_name + '.zip', quiet=False)
        zip = ZipFile(dataset_name + '.zip')
        zip.extractall(dataset_name)
        zip.close()

    def list_all(self):
        """
            list all urdf or xml file
        """
        for file in os.listdir(self.name):
            if file.find('.urdf') != -1 or file.find('.xml') != -1:
                print(file,'\n')

    def show_link(self, path):
        """
            show urdf file link
        """

        robot = URDF.load(path)

        for link in robot.links:
            print(link.name)